body {
  background-color: #ddd;
  color: #333;
  margin-top: 2em;
  margin-left: 5%;
  margin-right: 5%;
}

body, p, ol, ul, td {
  font-family: 'Lucida Sans Unicode', 'Lucida Grande', sans-serif;
  font-size:   13px;
  line-height: 18px;
}

div#login-header {
  margin: 1.5em;
}



  .tags ul{
    margin:1em 0;
    padding:.5em 10px;
    text-align:center;
    
    }
  .tags li{
    margin:0;
    padding:0;
    list-style:none;
    display:inline;

    }
  .tags li a{
    text-decoration:none;
    color:#fff;
    padding:0 2px;  
    }
  .tags li a:hover{ 
    color:#cff400;
    }
  

  
  /* alternative layout */

  .tags {
    text-align:left;
    padding:0;
    background:none;
    }
  .tags li{
    padding:2px 10px;
    background:#efefef;
    display:block;
    }
  
  .tag1{font-size:120%;}
  .tag2{font-size:120%;}
  .tag3{font-size:120%;}
  .tag4{font-size:120%;}
  .tag5{font-size:120%;}


  .tags .tag1{background:#7cc0f4;}
  .tags .tag2{background:#67abe0;}
  .tags .tag3{background:#4d92c7;}
  .tags .tag4{background:#3277ad;}
  .tags .tag5{background:#266ca2;}

#tags_container h2{
  font-size:200%; 
  text-decoration:underline; 
}

.container
{
	width: 100%;
	margin: 0 auto;

	padding-left: 2%;
	padding-right: 2%;
	#border: 1px solid #555;
}

.header
{
	width: 100%;
	margin-right: 0;
}

#login-header{
    color: #ddd;
    position: absolute;
    right: 10px;
    font-size: 11px;
    top: 5px;
}

#selectable .ui-selecting { background: #FECA40; }
#selectable .ui-selected { background: yellow; color: black; }
#selectable { list-style-type: none; margin: 0; padding: 0; width: 60%; }
#selectable li { margin: 3px; padding: 0.4em; font-size: 1.4em; height: 16px; }


.content{
	padding: 20px;
	background-color: #fff;
	-moz-box-shadow: 3px 3px 4px #aaa;
	-webkit-box-shadow: 3px 3px 4px #aaa;
	box-shadow: 3px 3px 4px #aaa;
	-moz-border-radius-topright: 4px;
  border-top-right-radius: 4px;

	-moz-border-radius-bottomright: 4px;
  border-bottom-right-radius: 4px;
  -moz-border-radius-bottomleft: 4px;
  border-bottom-left-radius: 4px;

  
  border: 1px solid #000000;
}

.sidebar {
   float: left;
  top:100px;
  left:5px;

  padding:5px;
  background-color:#eee;
  border:1px dashed #999;
  line-height:17px;
/* Again, the ugly brilliant hack. */
  voice-family: "\"}\"";
  voice-family:inherit;
  width:12%;

}

pre {
  background-color: #eee;
  padding: 10px;
  font-size: 11px;
}

a{
	text-decoration: none;
	color: #2c539e;
}

a:visited { color: #666; }
a:hover{
	text-decoration: underline;
}

	


/* ---- BUTTON CSS FROM: http://www.bloggerswatch.com/internet/css-trick-submit-button-should-look-same-everywhere/ */
.actions a, .actions button, .inline-actions a, .inline-actions button, .saved, .sidebar, .sidebar button{
    margin:0 7px 0 0;
    background-color:#f5f5f5;
    #border:1px solid #dedede;
    border-top:1px solid #eee;
    border-left:1px solid #eee;
    font-size:12px;
    line-height:130%;
    text-decoration:none;
    font-weight:bold;
    color:#565656;
    cursor:pointer;
    padding:6px; 
    border:1px dotted gray;
}



.disabled-actions a:hover, .disabled-actions input:hover{
   color: red;
}

.inline-actions a:hover, .actions a:hover, .actions input:hover, .sidebar input:hover{
    color: green;
}

.inline-actions a[disabled]:hover, .actions a[disabled]:hover, .actions input[disabled]:hover, .sidebar input[disabled]:hover{
    color: red;
}

.small-actions a, .small-actions button{
    margin:0 7px 0 0;
    background-color:#f5f5f5;
    border:1px solid #dedede;
    border-top:1px solid #eee;
    border-left:1px solid #eee;
    font-size:10px;
    line-height:130%;
    text-decoration:none;
    font-weight:bold;
    color:#565656;
    cursor:pointer;
    padding:6px; 
    border:1px dotted gray;
}

.actions input, .sidebar input{
        margin:0 7px 0 0;
    background-color:#f5f5f5;
    border:1px solid #dedede;
    border-top:1px solid #eee;
    border-left:1px solid #eee;
    font-size:12px;
    line-height:130%;
    text-decoration:none;
    font-weight:bold;
    color:#565656;
    cursor:pointer;
    padding:6px; 
    border:1px dotted gray;
}

.actions a, .small-actions a{
width: 100px;
}

.actions button, .small-actions a, .inline-actions button, button{
    width:auto;
    overflow:visible;
    padding:6px; /* IE6 */
}

.actions button img, .actions a img, .inline-actions button img, .inline-actions a img{
    margin:0 3px -3px 0 !important;
    padding:0;
    border:none;
    width:16px;
    height:16px;

}


div.field, div.actions {
  margin-bottom: 10px;
}

#notice {
  color: green;
}

.field_with_errors {
  padding: 2px;
  background-color: red;
  display: table;
}

#error_explanation {
  width: 450px;
  border: 2px solid red;
  padding: 7px;
  
  padding-bottom: 0;
  margin-bottom: 20px;
  background-color: #f0f0f0;
}

#error_explanation h2 {
  text-align: left;
  padding: 5px 5px 5px 15px;
  font-size: 12px;
  margin: -7px;
  margin-bottom: 0px;
  background-color: #c00;
  color: #fff;
}

#error_explanation ul li {
  font-size: 12px;
  list-style: square;
}

.hidden_select{
	visibility:hidden;
}

h1{
  font-size: 24px;
  font-family: 'Palatino Linotype', 'Book Antiqua', Palatino, serif;
  padding-bottom: 10px;
  color: 	#2B65EC;
}

h2{
  font-family: 'Palatino Linotype', 'Book Antiqua', Palatino, serif;
  margin: 10px;
  font-size: 18px;
}

h3{
  font-family: 'Palatino Linotype', 'Book Antiqua', Palatino, serif;
	font-size: 16px;
	padding: 3px;
	margin: 0px;
}

h4{
  font-family: Georgia, Serif;
	font-size: 14px;
	padding: 0px;
	margin: 3px;
	display: inline;
}

table
{
	border-collapse: collapse;
	/* border: 0px;*/
}

.indent_block{
	margin-left: 15px;
	width: 80%;
}

.summary_heading{
	width: 99%;
	  font-family: Georgia, Serif;
	background-color: #eee;
	border: 1px solid #ccc;
	font-size: 16px;
	font-weight: bold;
	color: #444;
	padding: 3px;
}

.listing_item{
	width: 100%;
	border: 1px solid #ccc;
	color: #333;
	padding: 4px;
}

.listing_item:hover{
	background-color: #ddd;
}

.listing_item h3{
padding: 8px;
}

.listing_item .admin_functions{
	width: 100%;
	text-align: right;
}

.listing_item .indent_block{
	width: 80%;
}

table#bottom_nav{
width: 80%;
}
td.calculated{
	background-color: #ffff99
}

.calculated_checkbox{
	display: inline;
}

.checkbox {
  width: 19px;
  height: 25px;
  padding: 0 5px 0 0;
  background: url(../images/checkbox.gif) no-repeat;
  display: block;
  clear: left;
  float: left;
}

.icon_explanation{
border: 1px solid #ccc;
padding: 8px;
}

#validation_message, #key_questions_validation_message{
	font-weight: bold;
	color: #000;
	padding: 5px;
}
.success_message{
	color: #10783B;
	border: 1px solid #187A27;
	padding: 5px;
}
.error_message{
	/* background-color: #ED4548; */
	border: 1px solid #ED4548;
	color: #8F2224;
	padding: 5px;
}

.odd, .odd.td{
	background: #eee;
	vertical-align: middle;
}

.even, .even.td{
    background:#dedede;
    vertical-align: middle;
}

.odd td.calculated{
	background: #ffffcc;
	border-top: 1px dotted #cccc66;	
}

.even td.calculated{
	background: #ffff99;
	border-top: 1px dotted #cccc33;
}

.hidden{
  display:none;
}

.visible{
  display:inline;
}

.toggle_display_link{
	text-decoration: none;
	font-size: 20px;
	color: green;
}
.toggle_hide_link{
	text-decoration: none;
	font-size: 20px;
	color: red;
}

a.active_tab {
    

  background-color: #ffffff;
  border: 1px solid #000000;
  border-bottom-width: 0px;
  padding: 2px 1em 2px 1em;
  text-decoration: none;
}

a.tab {
  background-color: #f0f0f0;
  border: 1px solid #000000;
  border-bottom-width: 0px;
  padding: 2px 1em 2px 1em;
  text-decoration: none;
  text-align: right;
}


table.list_table{
width: 100%;
background:#fff;
border:1px solid gray;
color:#fff;
font:normal 12px verdana, arial, helvetica, sans-serif;
margin-left: auto;
margin-right: auto;
}

table.list_table td{
padding: 10px;
text-align: center;
vertical-align:middle;
color:#363636;

}

table.list_table tr{
border:2px gray;

}

table.list_table tr:hover { 
 background:yellow;
 #border:1px solid #99cc99;
 #color:#000000;
}

table.non_centered_table td{
text-align: left;
#width: 60%;
}

table.list_table th{
background:#777;
border: 1px dotted #444;
 padding:3px 10px 3px 10px;
}

table.list_table tr th:hover { 
 /*background:#ccffcc;*/
 /*border:1px dotted gray;*/
 /*color:#fff;*/
}

table.complex_header th{
border: 1px dotted gray;
padding: 4px;
}

table#bottom_nav{
width: 100%;
}

table.form_table{
}

.actions{
text-align: right;
}

#top-header{
	margin-top: -8px;
	margin-left: -8px;
	padding-right: 8px;
	padding-left: 8px;
	width: 100%;
}

.header-sub{
margin-top: 50px;
width: 100%;
height: 40px;
padding-bottom:8px;
padding-left: 8px;
padding-right: 8px;
margin-left: -8px;
background-color: #ddd;
}

.login-header{
	width: 100%;
	margin: 0px;
	height: 24px;
	background-color: #aaa;
	font-color: black;
	position: absolute;
	top: 0px;
	left: 0px;
	border-bottom: 1px solid #888;
}

.login-link{
float:left;
margin-left: 8px;
margin-top: 3px;
color: #000;
font-weight: bold;
font-size: 10px;
font-family: verdana, sans-serif;
}

.login-link a, .login-link a:visited, .login-link a:active{
	text-transform: uppercase;
	font-weight: normal;
	text-decoration: none;
	color: #333;
  font-color: black;
}

.login-link a:hover {
	color: #000;
}

ul#menu {
margin-left: -40px;
padding-top: 8px;
}


ul#menu li {
  display: inline; /* Shows each item side-by-side */
  list-style-type: none; /* Gets rid of the bullet points */
  font-family: Georgia, Serif;
  font-size: 15px;
}

ul#menu li.title{
padding-right: 35px;
font-size:  25px;
font-family: 'Palatino Linotype', 'Book Antiqua', Palatino, serif;
  font-family: Georgia;
  #text-shadow: #999 2px 2px 3px;
  #font-weight: bold;
color: #333;
width: 700px;
}
 
ul#menu a {
 padding: 3px;
 padding-left: 10px;
 padding-right: 10px;
  color: #333;
  font-size: 12px;
  text-decoration: none;
}

ul#menu a:hover{
	  color: #888;
}

#login-form{
width: 240px;
height: 100px;
background-color: #ccc;
padding: 10px;
float: right;
-moz-border-radius-bottomright: 20px;
border-bottom-right-radius: 20px;
-moz-border-radius-bottomleft: 50px;
border-bottom-left-radius: 50px;
z-index: 100;
}

#login-form .form_table, #login-form .form_table td
{
font-size: 16px;
font-weight: bold;
font-family: Century Gothic, sans-serif;
color: #525252;
padding: 2px;
}

#login-form .form_table td.remember
{
font-size: 12px;
font-weight: bold;
}

#login-form .form_table td input
{
font-size: 14px;
border: 1px solid #aaa;
#border: 0px;
}

#login-form .form_table td input:focus, .searchbar input:focus
{
background: #D7F2F5;
color: #003333;
}


.breadcrumbs{
font-family: verdana, sans-serif;
margin-top: -20px;
margin-bottom: 15px;
color: #333;

font-size:10px;
	margin-left: -8px;
	margin-right: -8px;
	padding-right: 8px;
	padding-left: 8px;
}


.breadcrumbs a{

color: blue;
}

.breadcrumbs a:hover{
color: #999;
}

.loading{
    display: none;  
}

.searchbar{
	float:right;
	margin-top: 3px;
	margin-right: 5px;

.login-link a{
	text-decoration: underline;
	font-size: 11px;
	color: #eee;
}
.login-link a:visited{
	text-decoration: underline;
	font-size: 11px;
	color: #eee;
}
.login-link a:active{
	text-decoration: underline;
	font-size: 11px;
	color: #eee;
}
.login-link a:hover {
	color: #ccc;
}

.searchbar input
{
	background-color: #ccc;
	font-size: 12px;
	border: 0px;
}

input[type="text"]{
border: 0px;
}

.searchbar input:focus
{
background: #D7F2F5;
color: #003333;
}

table.pretty {
  margin: 1em 1em 1em 2em;
  background: whitesmoke;
  border-collapse: collapse;
}
table.pretty th, table.pretty td {
  border: 1px silver solid;
  padding: 0.2em;
}
table.pretty th {
  background: gainsboro;
  text-align: center;
}
table.pretty caption {
  margin-left: inherit;
  margin-right: inherit;
}


